@@ -1,3 +1,34 @@
+/* Copyright (c) 1993, 1994  Washington University in Saint Louis
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met: 1. Redistributions of source code must retain the above copyright
+ * notice, this list of conditions and the following disclaimer. 2.
+ * Redistributions in binary form must reproduce the above copyright notice,
+ * this list of conditions and the following disclaimer in the documentation
+ * and/or other materials provided with the distribution. 3. All advertising
+ * materials mentioning features or use of this software must display the
+ * following acknowledgement: This product includes software developed by the
+ * Washington University in Saint Louis and its contributors. 4. Neither the
+ * name of the University nor the names of its contributors may be used to
+ * endorse or promote products derived from this software without specific
+ * prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY WASHINGTON UNIVERSITY AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASHINGTON
+ * UNIVERSITY OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+ * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+ * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
+ */
+
 #include "config.h"
 
 #include <stdio.h>
@@ -17,7 +48,7 @@
 #include "pathnames.h"
 #include "extensions.h"
 
-#ifdef SVR4
+#if defined(SVR4) || defined(ISC)
 #include <fcntl.h>
 #endif
 
@@ -29,6 +60,8 @@ extern int nameserved,
   guest,
   use_accessfile;
 char Shutdown[MAXPATHLEN];
+#define MAXLINE	80
+static  char  incline[MAXLINE];
 int pidfd = -1;
 
 extern int fnmatch();
@@ -111,7 +144,8 @@ validtime(char *ptr)
             return (parsetime(ptr));
         *nextptr = '\0';
         good = parsetime(ptr);
-        *nextptr++ = '|';       /* gotta restore the | or things get skipped! */
+        /* gotta restore the | or things get skipped! */
+        *nextptr++ = '|';
         if (good)
             return (1);
         ptr = nextptr;
@@ -127,13 +161,40 @@ validtime(char *ptr)
 
 hostmatch(char *addr)
 {
-    if (addr == NULL)
-        return (0);
+    FILE  *incfile;
+    char  *ptr;
+    int   found = 0;
+
+    if (addr == NULL) return(0);
 
     if (isdigit(*addr))
-        return (fnmatch(addr, remoteaddr, NULL));
+        return(fnmatch(addr, remoteaddr, NULL));
+    else if (*addr == '/') {
+        /*
+         * read addrglobs from named path using similar format as addrglobs
+         * in access file
+         */
+        if ((incfile = fopen(addr, "r")) == NULL) {
+            if (errno != ENOENT)
+                syslog(LOG_ERR,
+                       "cannot open addrglob file %s: %s", addr,
+                       strerror(errno));
+            return(0);
+        }
+        
+        while (!found && (fgets(incline, MAXLINE, incfile) != NULL)) {
+            ptr = strtok(incline, " \t\n");
+            found = 1;
+            while (!found && ((ptr = strtok(NULL, " \t\n")) != NULL)) {
+                if (ptr && hostmatch(ptr))
+                    found = 1;
+            }
+        }
+        fclose(incfile);
+        return(found);
+    }
     else
-        return (fnmatch(addr, remotehost, FNM_NOCASE));
+        return(fnmatch(addr, remotehost, FNM_NOCASE));
 }
 
 /*************************************************************************/
@@ -491,11 +552,23 @@ acl_join(char *class)
     int which,
       avail;
     pid_t buf[MAXUSERS];
-	pid_t procid;
+    char pidfile[MAXPATHLEN];
+    pid_t procid;
 #ifndef HAVE_FLOCK
-	struct flock arg;
+    struct flock arg;
 #endif
 
+    /* 
+     * if pidfd was not opened previously... 
+     * pidfd must stay open after the chroot(~ftp)  
+     */
+
+    sprintf(pidfile, _PATH_PIDNAMES, class);
+
+    if (pidfd < 0) {
+        pidfd = open(pidfile, O_RDWR | O_CREAT, 0644);
+    }
+
     if (pidfd < 0) {
         syslog(LOG_ERR, "open of pid file failed: %s", strerror(errno));
         return;
@@ -505,16 +578,16 @@ acl_join(char *class)
     while (flock(pidfd, LOCK_EX)) {
         syslog(LOG_ERR, "sleeping: flock of pid file failed: %s",
 #else 
-	arg.l_type = F_WRLCK;
-	arg.l_whence = arg.l_start = arg.l_len = 0;
-	while ( -1 == fcntl( pidfd, F_SETLK, &arg) ) {
-		syslog(LOG_ERR, "sleeping: fcntl lock of pid file failed: %s",
+    arg.l_type = F_WRLCK;
+    arg.l_whence = arg.l_start = arg.l_len = 0;
+    while ( -1 == fcntl( pidfd, F_SETLK, &arg) ) {
+        syslog(LOG_ERR, "sleeping: fcntl lock of pid file failed: %s",
 #endif
                strerror(errno));
         sleep(1);
     }
 
-	procid = getpid();
+    procid = getpid();
 
     lseek(pidfd, 0, L_SET);
     if (read(pidfd, buf, sizeof(buf)) < sizeof(buf))
@@ -527,9 +600,15 @@ acl_join(char *class)
             avail = which;
             buf[which] = 0;
         } else if (buf[which] == procid) {
-			/* already exists in pid file... */
-			return;
-		}
+            /* already exists in pid file... */
+#ifdef HAVE_FLOCK
+            flock(pidfd, LOCK_UN);
+#else
+            arg.l_type = F_UNLCK; arg.l_whence = arg.l_start = arg.l_len = 0; 
+            fcntl(pidfd, F_SETLK, &arg);
+#endif
+            return;
+        }
     }
 
     buf[avail] = procid;
@@ -539,12 +618,95 @@ acl_join(char *class)
 #ifdef HAVE_FLOCK
     flock(pidfd, LOCK_UN);
 #else
-	arg.l_type = F_UNLCK; arg.l_whence = arg.l_start = arg.l_len = 0;
-	fcntl(pidfd, F_SETLK, &arg);
+    arg.l_type = F_UNLCK; arg.l_whence = arg.l_start = arg.l_len = 0;
+    fcntl(pidfd, F_SETLK, &arg);
 #endif
 
 }
 
+/*************************************************************************/
+/* FUNCTION  : acl_remove                                                */
+/* PURPOSE   : remove the current process to the list of processes in    */
+/*             the specified class.                                      */
+/* ARGUMENTS : The name of the class to remove                           */
+/*************************************************************************/
+
+void
+acl_remove()
+{
+    char class[1024];
+    int which,
+      avail;
+    pid_t buf[MAXUSERS];
+    char pidfile[MAXPATHLEN];
+    pid_t procid;
+#ifndef HAVE_FLOCK
+    struct flock arg;
+#endif
+
+    if (!acl_getclass(class)) {
+        return;
+    }
+
+    /* 
+     * if pidfd was not opened previously... 
+     * pidfd must stay open after the chroot(~ftp)  
+     */
+
+    sprintf(pidfile, _PATH_PIDNAMES, class);
+
+    if (pidfd < 0) {
+        pidfd = open(pidfile, O_RDWR | O_CREAT, 0644);
+    }
+
+    if (pidfd < 0) {
+        syslog(LOG_ERR, "open of pid file failed: %s", strerror(errno));
+        return;
+    }
+
+#ifdef HAVE_FLOCK
+    while (flock(pidfd, LOCK_EX)) {
+        syslog(LOG_ERR, "sleeping: flock of pid file failed: %s",
+#else 
+    arg.l_type = F_WRLCK;
+    arg.l_whence = arg.l_start = arg.l_len = 0;
+    while ( -1 == fcntl( pidfd, F_SETLK, &arg) ) {
+        syslog(LOG_ERR, "sleeping: fcntl lock of pid file failed: %s",
+#endif
+               strerror(errno));
+        sleep(1);
+    }
+
+    procid = getpid();
+
+    lseek(pidfd, 0, L_SET);
+    if (read(pidfd, buf, sizeof(buf)) < sizeof(buf))
+        for (which = 0; which < MAXUSERS; buf[which++] = 0)
+            continue;
+
+    avail = 0;
+    for (which = 0; which < MAXUSERS; which++) {
+        if ((buf[which] == 0) || (kill(buf[which], 0) == -1)) {
+            avail = which;
+            buf[which] = 0;
+        } else if (buf[which] == procid) {
+            buf[which] = 0;
+        }
+    }
+
+    lseek(pidfd, 0, L_SET);
+    write(pidfd, buf, sizeof(buf));
+#ifdef HAVE_FLOCK
+    flock(pidfd, LOCK_UN);
+#else
+    arg.l_type = F_UNLCK; arg.l_whence = arg.l_start = arg.l_len = 0;
+    fcntl(pidfd, F_SETLK, &arg);
+#endif
+
+    close(pidfd);
+    pidfd = -1;
+}
+
 /*************************************************************************/
 /* FUNCTION  : pr_mesg                                                   */
 /* PURPOSE   : Display a message to the user                             */
@@ -613,19 +775,20 @@ access_ok(int msgcode)
         return (1);
 
     if (aclbuf == NULL) {
-        syslog(LOG_INFO, "ACCESS DENIED (error reading access file) TO %s [%s]",
-               remotehost, remoteaddr);
+        syslog(LOG_NOTICE, 
+		"ACCESS DENIED (error reading access file) TO %s [%s]",
+                 remotehost, remoteaddr);
         return (0);
     }
     if (acl_deny(msgfile)) {
         pr_mesg(msgcode, msgfile);
-        syslog(LOG_INFO, "ACCESS DENIED (deny command) TO %s [%s]",
+        syslog(LOG_NOTICE, "ACCESS DENIED (deny command) TO %s [%s]",
                remotehost, remoteaddr);
         return (0);
     }
     /* if user is not in any class, deny access */
     if (!acl_getclass(class)) {
-        syslog(LOG_INFO, "ACCESS DENIED (not in any class) TO %s [%s]",
+        syslog(LOG_NOTICE, "ACCESS DENIED (not in any class) TO %s [%s]",
                remotehost, remoteaddr);
         return (0);
     }
@@ -636,7 +799,7 @@ access_ok(int msgcode)
         acl_join(class);
         return (1);
     } else {
-        syslog(LOG_INFO, "ACCESS DENIED (user limit %d; class %s) TO %s [%s]",
+        syslog(LOG_NOTICE, "ACCESS DENIED (user limit %d; class %s) TO %s [%s]",
                limit, class, remotehost, remoteaddr);
         pr_mesg(msgcode, msgfile);
         return (-1);
